#! /usr/bin/env node
"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const yargs = require("yargs");
const path_1 = require("path");
const fs_1 = require("fs");
const index_1 = require("../index");
const helper_1 = require("../helper");
const pretty_1 = require("../helper/pretty");
const cli_helper_1 = require("./cli.helper");
const stringify = (data) => JSON.stringify(data, undefined, 2) + '\n';
// Note: After TS transpilation, the following path refers to the directory: `./dist/src/package.json`
// (for details see the script: `npm run postbuild`).
const pkgVersion = require(path_1.join(__dirname, '../package.json')).version; // tslint:disable-line:no-var-requires
yargs
    .locale('en')
    .version(pkgVersion).alias('version', 'v')
    .help('help').alias('help', 'h')
    .usage('Usage: $0 <path> [options]')
    .demandCommand(1, 'Error: argument <path> is missing!')
    .option('all', {
    alias: 'a',
    describe: 'Include hidden files in output',
    type: 'boolean'
}).option('content', {
    alias: 'c',
    describe: 'Add files content to output',
    type: 'boolean'
}).option('depth', {
    alias: 'd',
    coerce: cli_helper_1.getDepthArg,
    describe: 'Maximum depth of directories',
    type: 'number',
    default: index_1.INFINITE_DEPTH // Setting a default value ensures that the coerce function will always be called
}).option('relative', {
    alias: 'r',
    describe: 'Use relative path',
    type: 'boolean'
}).option('include', {
    alias: 'i',
    describe: 'List of directory paths to include in output',
    type: 'array'
}).option('exclude', {
    alias: 'e',
    describe: 'List of directory paths to exclude from output',
    type: 'array'
}).option('glob', {
    alias: 'g',
    describe: 'Match files based on glob pattern',
    type: 'array'
}).option('sort', {
    alias: 's',
    coerce: cli_helper_1.getSortArg,
    describe: 'Sort output ' + cli_helper_1.getSortDesc(),
    type: 'number',
    default: 0
}).option('clean', {
    alias: 'n',
    describe: 'Clean empty directories from output',
    type: 'boolean'
}).option('flatten', {
    alias: 'f',
    describe: 'Flatten output',
    type: 'boolean'
}).option('pretty', {
    alias: 'p',
    coerce: cli_helper_1.getPrettyArg,
    describe: 'Pretty-print output',
    type: 'string'
}).option('watch', {
    alias: 'w',
    describe: 'Watch filesystem',
    type: 'boolean'
}).option('output', {
    alias: 'o',
    describe: 'Output file path',
    type: 'string'
}).option('debug', {
    describe: 'Add debugging information to output',
    type: 'boolean'
});
// process.stdout.write(stringify(yargs.argv)); // For debugging
const a = yargs.argv;
const path = a._[0];
const opts = {
    all: a.all,
    content: a.content,
    depth: a.depth,
    relative: a.relative,
    include: a.include || [],
    exclude: a.exclude || [],
    glob: a.glob || [],
    sort: a.sort
};
const helper = {
    clean: a.clean,
    flatten: a.flatten,
    pretty: a.pretty
};
const watchMode = a.watch;
const outputPath = a.output || undefined;
const debugMode = a.debug;
const handleTree = (tree) => {
    // Note that if the output is flattened there's no need to clean it!
    // Because the flatten version only contains the files (and not the directories).
    const output = helper.flatten ? helper_1.flatten(tree) :
        helper.clean ? helper_1.clean(tree) :
            tree;
    let outputStr;
    if (helper.pretty) {
        const render = pretty_1.renderer[helper.pretty];
        outputStr = pretty_1.pretty(output, render) + '\n';
    }
    else {
        outputStr = stringify(output);
    }
    if (debugMode) {
        const debug = {
            opts,
            path,
            helper,
            watchMode,
            output: helper.pretty ? outputStr : output,
            outputPath
        };
        process.stdout.write(stringify(debug));
    }
    else {
        if (outputPath) {
            fs_1.writeFile(path_1.resolve(outputPath), outputStr, (error) => {
                if (error)
                    throw error;
            });
        }
        else {
            process.stdout.write(outputStr);
        }
    }
};
if (path) {
    const treeview = new index_1.TreeView(opts);
    if (!watchMode) {
        treeview.process(path).then(handleTree).catch(cli_helper_1.exit);
    }
    else {
        treeview.on('ready', handleTree).on('tree', handleTree).watch(path);
    }
}
