"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const fs_1 = require("fs");
const path_1 = require("path");
const minimatch = require("minimatch");
const Model = require("./model");
const binary_1 = require("./helper/binary");
const watch_1 = require("./watch");
exports.INFINITE_DEPTH = -1;
class TreeView {
    constructor(opts) {
        this.opts = {
            all: false,
            content: false,
            depth: exports.INFINITE_DEPTH,
            relative: false,
            include: [],
            exclude: [],
            glob: [],
            sort: Model.Sorting.Alpha
        };
        this.events = new events_1.EventEmitter();
        this.inject();
        Object.assign(this.opts, opts || {});
        this.formatOpts();
    }
    static addTime(item, stats) {
        item.created = stats.birthtime;
        item.modified = stats.mtime;
    }
    static removeHidden(pathfiles) {
        return pathfiles.filter(pathfile => (pathfile.match(/[^\/|\\]+$/) || [''])[0][0] !== '.');
    }
    static getMainPaths(paths) {
        const main = paths.sort().reduce((acc, curr, index) => {
            if (!index || !curr.startsWith(acc.prev)) {
                acc.paths.push(curr);
                acc.prev = curr;
            }
            return acc;
        }, {
            paths: [],
            prev: ''
        });
        return main.paths;
    }
    on(event, listener) {
        this.events.addListener(event, listener);
        return this;
    }
    removeListeners(event) {
        this.events.removeAllListeners(event);
        return this;
    }
    process(path, cb) {
        const rootPath = this.providers.resolve(path);
        const ctx = {
            rootPath,
            getPath: this.getPathFactory(rootPath),
            path: rootPath,
            depth: 0
        };
        const promise = this.discover(ctx);
        promise.then(tree => this.lastResult = { rootPath, tree }, error => error);
        if (cb)
            promise.then(tree => cb(null, tree), error => cb(error));
        return promise;
    }
    watch(path, watchFn = watch_1.default) {
        const rootPath = this.providers.resolve(path);
        let pathsStack = [];
        let ready = false;
        const refresh = () => {
            if (pathsStack.length) {
                ready = false;
                this.refreshResult(pathsStack).then((tree) => {
                    this.emit('tree', tree);
                    refresh();
                });
                pathsStack = [];
            }
            else {
                ready = true;
            }
        };
        const watcher = watchFn(rootPath, (fullpaths) => {
            pathsStack.push(...fullpaths);
            if (ready) {
                refresh();
            }
        });
        Promise.all([this.process(rootPath), watcher.ready]).then(([tree]) => {
            this.emit('ready', tree);
            refresh();
        });
        return { close: watcher.close };
    }
    refreshResult(paths) {
        const filtered = this.filterResult([].concat(paths));
        if (!this.opts.all) {
            filtered.remains = TreeView.removeHidden(filtered.remains);
        }
        return Promise.all(TreeView.getMainPaths(filtered.remains).map(pathname => this.extendResult(pathname))).then(() => Promise.all(filtered.matchs.map(match => this.updateResult(match)))).then(() => this.lastResult.tree);
    }
    inject() {
        this.providers = { join: path_1.join, resolve: path_1.resolve, relative: path_1.relative, readFile: fs_1.readFile, readdir: fs_1.readdir, stat: fs_1.stat };
    }
    formatOpts() {
        [this.opts.include, this.opts.exclude].forEach(paths => paths.forEach((path, n) => paths[n] = this.providers.resolve(path)));
    }
    getPathFactory(rootPath) {
        return (item) => this.providers.resolve(this.opts.relative ? rootPath : '', item.path, item.name);
    }
    discover(ctx, tree = []) {
        return new Promise((success, reject) => {
            this.providers.readdir(ctx.path, (error, files) => {
                if (error) {
                    reject(error);
                    return;
                }
                if (!this.opts.all) {
                    files = TreeView.removeHidden(files);
                }
                files = files.filter(file => !tree.some(item => item.name === file)); // Prevent duplicate scan
                if (!files.length) {
                    success(tree);
                    return;
                }
                const tasks = files.map(name => this.getTreeNode(ctx, name));
                Promise.all(tasks).then((items) => {
                    items.forEach((item) => { if (item)
                        tree.push(item); });
                    success(this.sort(tree));
                });
            });
        });
    }
    getTreeNode(ctx, name) {
        // `ctx.path` and `pathname` are always absolute...
        const pathname = this.providers.resolve(ctx.path, name);
        // ...while `itemPath` and `itemPathname` are relative when the option `relative` is `true`.
        const itemPath = this.opts.relative ? this.providers.relative(ctx.rootPath, ctx.path) : ctx.path;
        const itemPathname = this.opts.relative ? this.providers.join(itemPath, name) : pathname;
        const item = {
            name,
            path: itemPath,
            pathname: itemPathname,
            depth: ctx.depth
        };
        const pathfile = ctx.getPath(item);
        return new Promise((success) => {
            this.providers.stat(pathfile, (err, stats) => {
                if (err) {
                    this.addError(ctx, item, err);
                    success(item);
                }
                else {
                    TreeView.addTime(item, stats);
                    let task;
                    let checked = false;
                    const cb = () => checked ? success(item) : success();
                    // The glob pattern should match "absolute path" when the
                    // option `relative` is `false` and "relative path" otherwise.
                    // Accordingly, we check the file against `item.pathname` (and not `pathname`).
                    // On the other hand, we check the directory against `path` or `pathname` which are always absolute.
                    if (stats.isFile() && this.checkFile(item.pathname, ctx) && this.checkDirectory(ctx.path, ctx, true)) {
                        task = this.addFile(ctx, item, stats);
                        checked = true;
                    }
                    else if (stats.isDirectory() && this.checkDirectory(pathname, ctx)) {
                        task = this.addDir(ctx, item);
                        checked = true;
                    }
                    task ? task.then(cb) : cb();
                }
            });
        });
    }
    checkFile(file, search = {}) {
        const globList = [...this.opts.glob, ...(search.glob || [])];
        const globCb = (glob) => minimatch(file, glob);
        return !globList.length || globList.findIndex(globCb) !== -1;
    }
    checkDirectory(dir, search = {}, strict = false) {
        const incList = [...this.opts.include, ...(search.include || [])];
        const excList = [...this.opts.exclude, ...(search.exclude || [])];
        const incCb = strict
            ? (path) => dir.startsWith(path) // For a file
            : (path) => dir.length > path.length ? dir.startsWith(path) : path.startsWith(dir); // For a dir
        const included = !incList.length || incList.findIndex(incCb) !== -1;
        const excluded = excList.includes(dir);
        return included && !excluded;
    }
    addError(ctx, item, error) {
        item.error = error;
        this.emit('item', item, ctx);
    }
    addFile(ctx, item, stats) {
        item.type = 'file';
        item.size = stats.size;
        item.ext = path_1.extname(item.name).slice(1); // remove '.'
        item.binary = binary_1.isBinaryPath(item.name);
        this.emit('item', item, ctx);
        if (this.opts.content) {
            return this.addContent(ctx, item);
        }
        return null;
    }
    addContent(ctx, item) {
        return new Promise(success => this.providers.readFile(ctx.getPath(item), {
            encoding: item.binary ? 'base64' : 'utf8'
        }, (error, data) => {
            if (error) {
                item.error = error;
                item.content = '';
            }
            else {
                item.content = data.toString();
            }
            success();
        }));
    }
    addDir(ctx, item) {
        item.type = 'dir';
        item.nodes = [];
        this.emit('item', item, ctx);
        if (this.opts.depth === exports.INFINITE_DEPTH || ctx.depth < this.opts.depth) {
            const newCtx = Object.assign({}, ctx, { path: ctx.getPath(item), depth: ctx.depth + 1 });
            return this.discover(newCtx, item.nodes)
                .catch((error) => {
                item.error = error;
                return Promise.resolve(); // Don't break the walk...
            });
        }
        return null;
    }
    sort(tree) {
        if (this.opts.sort === Model.Sorting.Alpha) {
            return tree.sort((a, b) => a.name > b.name ? 1 : -1);
        }
        else {
            const x = this.opts.sort === Model.Sorting.FileFirst ? 1 : -1;
            return tree.sort((a, b) => {
                if (a.type === 'dir' && b.type === 'file')
                    return x;
                if (a.type === 'file' && b.type === 'dir')
                    return -x;
                if (a.error && !b.error)
                    return 1;
                if (!a.error && b.error)
                    return -1;
                return a.name > b.name ? 1 : -1;
            });
        }
    }
    emit(event, ...args) {
        this.events.emit(event, ...args);
        this.events.emit('all', event, ...args);
    }
    filterResult(paths) {
        const remains = paths.map(this.opts.relative
            ? p => this.providers.relative(this.lastResult.rootPath, p)
            : p => this.providers.resolve(p));
        const filter = (nodes) => {
            return nodes.reduce((acc, item) => {
                const index = remains.indexOf(item.pathname);
                if (index !== -1) {
                    acc.push({ item, parentNodes: nodes });
                    remains.splice(index, 1);
                    if (!remains.length) {
                        return acc;
                    }
                }
                if (item.type === 'dir') {
                    for (const pathname of remains) {
                        if (pathname.startsWith(item.pathname)) {
                            acc = acc.concat(filter(item.nodes));
                            break;
                        }
                    }
                }
                return acc;
            }, []);
        };
        return { matchs: filter(this.lastResult.tree), remains };
    }
    updateResult(match) {
        // Remember! `ctx.path` is always absolute!
        const ctxPath = this.providers.resolve(this.opts.relative ? this.lastResult.rootPath : '', match.item.path);
        return new Promise((success) => {
            const ctx = {
                rootPath: this.lastResult.rootPath,
                getPath: this.getPathFactory(this.lastResult.rootPath),
                path: ctxPath,
                depth: match.item.depth
            };
            this.getTreeNode(ctx, match.item.name).then((item) => {
                const index = match.parentNodes.indexOf(match.item);
                if (index !== -1) {
                    if (item && !item.error) {
                        match.parentNodes.splice(index, 1, item); // Replace
                        this.emit('change', item);
                    }
                    else {
                        match.parentNodes.splice(index, 1); // Remove
                        this.emit('unlink', match.item);
                    }
                }
                else if (item && !item.error) {
                    match.parentNodes.push(item); // Add
                    this.sort(match.parentNodes);
                    this.emit('add', item);
                }
                success();
            });
        });
    }
    extendResult(pathname) {
        const relPathname = this.opts.relative
            ? pathname
            : this.providers.relative(this.lastResult.rootPath, pathname);
        const names = relPathname.split(/\/|\\/g);
        let tree = this.lastResult.tree;
        let parent = null;
        let name;
        // tslint:disable-next-line:no-conditional-assignment
        while (name = names.shift()) {
            const match = tree.find(item => item.name === name);
            if (match && match.type === 'dir') {
                parent = match;
                tree = match.nodes;
            }
            else {
                break;
            }
        }
        if (name) {
            const parentNodes = parent ? parent.nodes : this.lastResult.tree;
            const depth = parent ? parent.depth + 1 : 1;
            const path = parent ? parent.pathname : (this.opts.relative ? '' : this.lastResult.rootPath);
            const item = {
                name,
                path,
                pathname: this.providers.join(path, name),
                depth
            };
            return this.updateResult({ item, parentNodes });
        }
        return Promise.resolve();
    }
}
exports.TreeView = TreeView;
